% Copyright 2019. All Rights Reserved
% Code by Shuang Li
% For Paper, "The Landscape of Non-convex Empirical Risk with Degenerate Population Risk"
% Advances in Neural Information Processing Systems, 2019
% by Shuang Li, Gongguo Tang, and Michael B. Wakin
% This code is used to reproduce Figure 5.

%% To compute local minima and saddles, please run the Newton's method separately with different initializations.

clc; clear; close all;

%%

M = 10;     % number of samples
N = 2;     % signal dimension
bd = 2;    
x1 = -bd:bd/100:bd;
x2 = -bd:bd/100:bd;
[X1,X2] = meshgrid(x1,x2);

% ---------------- compute population risk -----------------

g = ((X1.*X1-1).^2+(X2.*X2-1).^2+2*(X1.*X2+1).^2)+ 0.5*(X1.^2+X2.^2-2).^2;  % population risk

% ----------------- compute empirical risk -----------------

f = zeros(size(g));  % empirical risk
A = zeros(N,M);      % sensing matrices
for m = 1:M
    Am = randn(N,1);
    A(:,m) = Am;
    f = f + ((abs(Am(1)*X1+Am(2)*X2)).^2 - (abs(Am(1)-Am(2))).^2).^2;
end
f = f/2/M;

%% ------------------ compute global minima ------------------

fmin = min(min(f));
[indf1, indf2] = find(f == fmin);
xf1 = x1(indf1);
xf2 = x2(indf2);

%% -- compute local minima and saddles with Newton's method --

x1n0 = 0.8;        
x2n0 = 0.4;   
xn = [x1n0; x2n0];     % choose initialization point
xs = [1;-1];           % global mimium
gradf = [1 0];         % initialize gradient
count = 0;
while norm(gradf) >= 1e-10 
    % compute gradient
    gradf = zeros(N,1);
    for m = 1:M
        Am = A(:,m);
        gradf = gradf + Am*(xn'*Am)^3 - Am*(xn'*Am)*(xs'*Am)^2;
    end
    gradf = gradf*2/M;
    % compute Hessian
    hessf = zeros(N,N);
    for m = 1:M
        Am = A(:,m);
        hessf = hessf + 3*(Am*Am')*(xn'*Am)^2 - Am*Am'*(xs'*Am)^2;
    end
    hessf = hessf*2/M;
    % update xn
    xn = xn - pinv(hessf)*gradf;
    
end

% the final xn can be local/global minima, or saddle points



%%
figure(1)
fs = 30;
ms = 20;
numline = 100;
contourf(x1,x2,g,numline,'ShowText','off'); hold on;
colormap cool;
plot(-1,1,'r*','MarkerSize',ms);hold on;
plot(1,-1,'r*','MarkerSize',ms);hold on;
plot(1/sqrt(3),1/sqrt(3),'r+','MarkerSize',ms);hold on;
plot(-1/sqrt(3),-1/sqrt(3),'r+','MarkerSize',ms);hold on;
plot(0,0,'r+','MarkerSize',ms);hold off;
xlabel('$x_1$','Interpreter','LaTex','FontSize',fs)
ylabel('$x_2$','Interpreter','LaTex','FontSize',fs)
set(gca,'FontSize',fs);
set(gcf, 'Color', 'w'); 


figure(2)
contourf(x1,x2,f,numline,'ShowText','off'); hold on;
colormap cool;
plot(-1,1,'r*','MarkerSize',ms);hold on;
plot(1,-1,'r*','MarkerSize',ms);hold on;
plot(1/sqrt(3),1/sqrt(3),'r+','MarkerSize',ms);hold on;
plot(-1/sqrt(3),-1/sqrt(3),'r+','MarkerSize',ms);hold on;
plot(0,0,'r+','MarkerSize',ms);hold on;
%plot(xn(1),xn(2),'bd','MarkerSize',ms+5);hold on;    % plot saddle points
%plot(xn(1),xn(2),'bo','MarkerSize',ms);hold on;      % plot local minima
plot(xf1,xf2,'bs','MarkerSize',ms+5);hold off;
xlabel('$x_1$','Interpreter','LaTex','FontSize',fs)
ylabel('$x_2$','Interpreter','LaTex','FontSize',fs)
set(gca,'FontSize',fs);
set(gcf, 'Color', 'w'); 



